/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/

/*
Code contains An Algorithm for Automatically Fitting Digitized Curves
by Philip J. Schneider
from "Graphics Gems", Academic Press, 1990
*/

#ifndef BEZIER_CUBIC_H
#define BEZIER_CUBIC_H

#include "BezierHelper.h"

namespace sofa {

namespace bezier4D {


class BezierCubic {
public:
    BezierCubic(vector<CurvePoint4>& _p, vector<double>& _u, vector<Point4>& _cp, vector<Point4>& _hats, double _l, bool _verbose) {
        curvePoints = _p;
        curvePointsParam = _u;
        controlPoints = _cp;
        tangents = _hats;
        length = _l;
        verbose = _verbose;
        tangentFactor = 0.001;
    }

    BezierCubic(bool _verbose) {
        length = 0.0;
        verbose = _verbose;
        tangentFactor = 0.001;
    }

    vector<CurvePoint4> curvePoints;
    vector<double> curvePointsParam;
    vector<Point4> controlPoints;
    vector<Point4> tangents;
    vector<double> segParams;

    double tangentFactor;

    double length;

    bool verbose;

    Point4 getPoint(double t);
    void GenerateBezier(vector<CurvePoint4> &cnodes,int first, int last, vector<double> params, Vector4 tHat1, Vector4 tHat2);
    double computeLength(size_t numPoints);

    void ChordLengthParameterize(vector<CurvePoint4> &p,size_t first, size_t last, vector<double>& u);
    double ComputeMaxError(vector<CurvePoint4>& cnodes, int first, int last, vector<double> params, int& splitPoint);

    void Reparameterize(vector<CurvePoint4>& p, int first,int last, vector<double>& params, vector<double>& newParams);
    double NewtonRaphsonRootFind(Point4 P, double params);

    //bool getRigidPoint(RigidCurvePoint4& _rigidPoint, double _t);
    bool getRigidPoint(RigidCurvePoint4& _rigidPoint, double _t, const RigidCurvePoint4& _rigidPointPrevious = RigidCurvePoint4());

    void clear() {
        curvePoints.clear();
        curvePointsParam.clear();
        controlPoints.clear();
        tangents.clear();
        length=0.0;
    }


    //void draw(DrawInfoMap &dim);
};

}

}


#endif
