#ifndef BEZIER_HELPER_H
#define BEZIER_HELPER_H

#include <iostream>
#include <string>
#include <string.h>
#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <iomanip>
#include <vector>
#include <iostream>
#include <istream>
#include <ostream>
#include <iterator>
#include <sstream>
#include <algorithm>
#include <map>
#include <math.h>
#include <set>
#include <stdio.h>
#include <sys/stat.h>

namespace sofa {

namespace bezier4D {


//#define ABS(a)		(((a)<0) ? -(a) : (a))
//#define ROUND(a)	((a)>0 ? (int)((a)+0.5) : -(int)(0.5-(a)))
//#define ZSGN(a)		(((a)<0) ? -1 : (a)>0 ? 1 : 0)
//#define SGN(a)		(((a)<0) ? -1 : 1)
//#define ASSERT(x) if (!(x)) fprintf(stderr," Assert failed: x\n");
#define _SQR(a)		((a)*(a))

//#define PI		3.141592	/* the venerable pi */
//#define PITIMES2	6.283185	/* 2 * pi */
//#define PIOVER2		1.570796	/* pi / 2 */
//#define E		2.718282	/* the venerable e */
//#define _SQRT2		1.414214	/* sqrt(2) */
//#define _SQRT3		1.732051	/* sqrt(3) */
//#define GOLDEN		1.618034	/* the golden ratio */
//#define DTOR		0.017453	/* convert degrees to radians */
//#define RTOD		57.29578	/* convert radians to degrees */

//#define SOFF 1000
//#define VOFF 1000000

using namespace std;

class Point4;
class Point3;

Point4 frame2Quater(Point3 vT, Point3 vB, Point3 vN);

struct MP_ID {
    MP_ID() {
        sID = -1;
        pID = -1;
    }

    MP_ID(int si, int pi) {
        sID = si;
        pID = pi;
    }

    int sID, pID;
};

class Point3 : public vector<double> {
public:
    Point3() {
        resize(3,0.0);
    }

    Point3(double _x, double _y, double _z) {
        resize(3,0.0);
        at(0) = _x;
        at(1) = _y;
        at(2) = _z;
    }

    void normalize() {
        double l = sqrt(_SQR(at(0))+_SQR(at(1))+_SQR(at(2)));
        for (size_t i = 0; i < 3; i++)
            at(i) /= l;
    }

    Point3 normalized() {
        double l = sqrt(_SQR(at(0))+_SQR(at(1))+_SQR(at(2)));
        Point3 r(at(0)/l, at(1)/l, at(2)/l);
        return r;
    }

    void set(double _x = 0.0, double _y = 0.0, double _z = 0.0) {
        at(0) = _x;
        at(1) = _y;
        at(2) = _z;
    }        

    double dist2(Point3 p) {
        return _SQR(p[0] - at(0)) + _SQR(p[1] - at(1)) + _SQR(p[2] - at(2));
    }


    double norm2() const {
        return(_SQR(at(0)) + _SQR(at(1)) + _SQR(at(2)));
    }

    double norm() const {
        return(sqrt(_SQR(at(0)) + _SQR(at(1)) + _SQR(at(2))));
    }

    void print() {
        cout << at(0) << " " << at(1) << " " << at(2);
    }

    double dot(Point3& _v) {
        return(at(0) * _v[0] + at(1) * _v[1] + at(2) * _v[2]);
    }


};
typedef Point3 Vector3;


Point3 cross(Point3 u, Point3 v);
double dot(Point3 u, Point3 v);


class Point4 : public vector<double> {
public:
    Point4() {
        resize(4,0.0);
    }

    Point4(double x, double y, double z, double scale = 1.0) {
        resize(4,0.0);
        at(0) = scale*x;
        at(1) = scale*y;
        at(2) = scale*z;
        at(3) = 0.0;
    }

    Point4(double x, double y, double z, double v, double scale = 1.0) {
        resize(4,0.0);
        at(0) = scale*x;
        at(1) = scale*y;
        at(2) = scale*z;
        at(3) = scale*v;
    }

    void print(string label="P4: ") {
        cout << label << " " << at(0) << " " << at(1) << " " << at(2) << " " << at(3) << endl;
    }

    void set(double x = 0.0, double y = 0.0, double z = 0.0, double w = 0.0, double scale = 1.0) {
        resize(4,0.0);
        at(0) = scale*x;
        at(1) = scale*y;
        at(2) = scale*z;
        at(3) = scale*w;
    }

    double norm2() const {
        return(_SQR(at(0)) + _SQR(at(1)) + _SQR(at(2)) + _SQR(at(3)));
    }

    double norm() const {
        return(sqrt(_SQR(at(0)) + _SQR(at(1)) + _SQR(at(2))) + _SQR(at(3)));
    }


    friend ostream& operator<<(ostream &_of, Point4& _p) {
        _of << _p[0] << " " << _p[1] << " " << _p[2] << " " << _p[3] << " ";
        return(_of);
    }

};
typedef Point4 Vector4;


class CurvePoint4 : public Point4 {
public:
    char type;   /// type of the node: S,E,C,s,e,i    
    int vid;     /// vessel ID (from the graph-algo)

    /// special data for topological points (S,E,s,e,C)
    vector<size_t> succ;        /// successor in branching point
    int pred;                   /// predecessor in branching point
    int topoID;             /// uniq ID for a key point

    /// no more used:
    //long int hashID;
    //vector<MP_ID> mpIdSucc;
    //MP_ID mpIdPred;
    //double param;

    CurvePoint4(double x, double y, double z, double v, double scale = 1.0) : Point4(x,y,z,v,scale) {
        type = 'u';
        pred = -1;
        //param = -1.0;
        topoID = -1.0;
        //pid = sid = -1;

    }

    CurvePoint4() : Point4() {
        type = 'u';
        pred = -1;
        //param = -1.0;
        topoID = -1.0;
        //pid = sid = -1;
        //vid = -1;
    }

    CurvePoint4(Point4 _p) : Point4(_p) {
        type = 'u';
        pred = -1;
        //param = -1.0;
        topoID = -1.0;
        //pid = sid = -1;
        //vid = -1;
    }

    void vID(int _vid) {
        vid = _vid;
        //hashID = (vid != -1 && sid != -1 && pid != -1) ? VOFF*vid + SOFF*sid + pid : -1;
    }
    int vID() { return vid; }
    
    bool isTopological() {
        return (type == 'S' || type == 's' || type == 'E' || type == 'e' || type == 'C');
    }

    //void sID(int _sid) {
    //    sid = _sid;
        //hashID = (vid != -1 && sid != -1 && pid != -1) ? VOFF*vid + SOFF*sid + pid : -1;
    //}
    //int sID() { return sid; }

    //void pID(int _pid) {
    //    pid = _pid;
        //hashID = (vid != -1 && sid != -1 && pid != -1) ? VOFF*vid + SOFF*sid + pid : -1;
    //}

    //int pID() { return pid; }

    friend ostream& operator<<(ostream &_of, CurvePoint4& _p) {
        _of << _p[0] << " " << _p[1] << " " << _p[2] << " " << _p[3] << " " << _p.type << " " << _p.vid;
        return(_of);
    }

};

class RigidCurvePoint4 : public CurvePoint4 {
public:
    Point4 matchingPosition;
    bool hasMatchingPosition;

    int indexInMO;

    Point4 quat;
    Point3 TV, NV, BV;
    double distanceFromStart;

    /// part for mapping
    Point4 baryCoord;
    int mappedTetra;

    /// \ part for mapping

    RigidCurvePoint4() : CurvePoint4() {
        hasMatchingPosition = false;
        distanceFromStart = -1.0;
        mappedTetra = -1;

        TV.set();
        BV.set();
        NV.set();

        quat.set();
        indexInMO = -1;
    }

    RigidCurvePoint4(CurvePoint4 _p, Point3 _t, double _distanceFromStart = -1.0) : CurvePoint4(_p) {
        distanceFromStart = _distanceFromStart;

        TV=_t.normalized();

        BV.set(0.0,1.0,0.0);
        NV.set(0.0,0.0,1.0);

        NV=cross(TV,BV);
        NV.normalize();
        BV=cross(NV,TV);
        BV.normalize();

        quat = frame2Quater(TV, BV, NV);
        mappedTetra = -1;

        indexInMO = -1;
    }

    RigidCurvePoint4(vector<string> tokens, double scale = 1.0) {
        distanceFromStart = atof(tokens[0].c_str());
        resize(4,0.0);
        for (size_t i = 0; i < 4; i++)
            at(i) = scale * atof(tokens[i+1].c_str());

        quat.resize(4,0.0);
        for (size_t i = 0; i < 4; i++)
            quat[i] = atof(tokens[i+5].c_str());

        matchingPosition.resize(4,0.0);
        if (tokens[9] == "R") {
            for (size_t i = 0; i < 4; i++)
                matchingPosition[i]=scale*atof(tokens[i+10].c_str());
            hasMatchingPosition = true;
        } else
            hasMatchingPosition = false;
        pred=atoi(tokens[14].c_str());
        /*int numSucc = atoi(tokens[15].c_str());
        succ.resize(numSucc);
        for (size_t i = 0; i < numSucc; i++)
            succ[i] = atoi(tokens[i+16].c_str());*/

        mappedTetra = -1;

        indexInMO = -1;

    }

    //bool mapPointToTetraVertices(ChMesh *pMesh, size_t el, Point4 &bc);

    void set(CurvePoint4 _p, Vector4 _t, const RigidCurvePoint4& _rigidPointPrevious = RigidCurvePoint4(), double _distanceFromStart = -1.0) {
        resize(4,0.0);
        for (size_t i = 0; i < 4; i++)
            at(i) = _p[i];
        type = _p.type;
        vid = _p.vid;
        //sid = _p.sid;
        //pid = _p.pid;
        //hashID = _p.hashID;
        //param = _p.param;
        succ = _p.succ;
        pred = _p.pred;
        distanceFromStart = _distanceFromStart;

        Vector3 _t3(_t[0], _t[1], _t[2]);

        TV=_t3.normalized();

        if (_rigidPointPrevious.BV.norm2() == 0.0) {  /// implicit parameter, orientation not initialized
            BV.set(0.0,1.0,0.0);
            if (fabs(dot(BV, TV) - 1.0) < 1e-5 ) {
                std::cout << "Two orientations too close, choosing orthogonal" << std::endl;
                BV.set(1.0,0.0,0.0);
            }            
        } else
            BV = _rigidPointPrevious.BV;

        //std::cout << "DOT: " << TV.dot(BV) << std::endl;
        NV=cross(TV,BV);
        NV.normalize();        
        BV=cross(NV,TV);
        BV.normalize();

        //std::cout << "TV: " << TV[0] << " " << TV[1] << " " << TV[2] << std::endl;
        //std::cout << "BV: " << BV[0] << " " << BV[1] << " " << BV[2] << std::endl;
        //std::cout << "NV: " << NV[0] << " " << NV[1] << " " << NV[2] << std::endl;

        quat = frame2Quater(TV, BV, NV);

        //std::cout << "Q = " << quat << std::endl;
    }

    void setMatchingPosition(RigidCurvePoint4 _rp) {
        for (size_t i = 0; i < 4; i++) {
            hasMatchingPosition = true;
            matchingPosition[i] = _rp[i];
        }
    }


    /*qglviewer::Vec getGLMatchingPos() {
        qglviewer::Vec r;
        for (size_t i = 0; i < 3; i++)
            r[i] = matchingPosition[i];
        return r;
    }*/

    //void draw(DrawInfoMap& dim);

    friend ofstream& operator<<(ofstream &_of, RigidCurvePoint4& _p) {
        _of << " " << _p.distanceFromStart;
        for (size_t i = 0; i < 4; i++)
            _of << " " << _p[i];
        for (size_t i = 0; i < 4; i++)
            _of << " " << _p.quat[i];
        _of << ( (_p.hasMatchingPosition) ? " R" : " N");

        for (size_t i = 0; i < 4; i++)
            _of << " " << _p.matchingPosition[i];

        _of << " " << _p.pred;
        _of << " " << _p.succ.size();
        for (size_t i = 0; i < _p.succ.size(); i++)
            _of << " " << _p.succ[i];

        _of << endl;

        return _of;
    }

    friend ostream& operator<<(ostream &_of, RigidCurvePoint4& _p) {
        _of << " " << _p.distanceFromStart;
        for (size_t i = 0; i < 4; i++)
            _of << " " << _p[i];
        for (size_t i = 0; i < 4; i++)
            _of << " " << _p.quat[i];
        _of << ( (_p.hasMatchingPosition) ? " R" : " N");

        for (size_t i = 0; i < 4; i++)
            _of << " " << _p.matchingPosition[i];

        _of << " " << _p.pred;
        _of << " " << _p.succ.size();
        for (size_t i = 0; i < _p.succ.size(); i++)
            _of << " " << _p.succ[i];

        _of << " " << _p.mappedTetra;

        _of << endl;

        return _of;
    }

};



/// functions

Point3 getDiff4(Point4 a, Point4 b);

double B0(double u);
double B1(double u);
double B2(double u);
double B3(double u);

Vector4 ComputeLeftTangent(Point4 p1, Point4 p2);
Vector4 ComputeRightTangent(Point4 p1, Point4 p2);
Vector4 ComputeCenterTangent(Point4 p1, Point4 p2, Point4 p3);

double V4SquaredLength(Vector4 a);
double V4Length(Vector4 a);

Vector4 V4Negated(Vector4 v);
void V4Negate(Vector4& v);

Vector4 V4Normalized(Vector4 v);
void V4Normalize(Vector4& v);

Vector4 V4ImposedLength(Vector4 v,double newlen);
void V4ImposeLength(Vector4& v,double newlen);

void V4Scale3(Vector4& v, double factor);
Vector4 V4Scaled3(Vector4& v, double factor);

Vector4 V4Add(Vector4 a,Vector4 b);
Vector4 V4Sub(Vector4 a,Vector4 b);

double V4Dot(Vector4 a,Vector4 b);
double V4Dist(Point4 a,Point4 b);
double V4Dist3(Point4 a,Point4 b);

Point4 BezierII(int degree, vector<Point4>&V, double t);

}

}


#endif


