import Sofa
import math
import yaml
import copy
import numpy as np
import random

from CellSimuOptions import CellData, FiloData

__file = __file__.replace('\\', '/') # windows

def createScene(root_node):    
    # root_node.createObject('RequiredPlugin', pluginName='ImageMeshAux')        
    # root_node.createObject('RequiredPlugin', pluginName='SofaCellSim')
    root_node.createObject('RequiredPlugin', pluginName='SofaBezierSkeletonization')
    root_node.createObject('RequiredPlugin', pluginName='SofaPardisoSolver')    
    root_node.createObject('PythonScriptController', name='TwoCells', filename=__file, classname='TwoCellsOEXID319Colliding')


class TwoCellsOEXID319Colliding (Sofa.PythonScriptController):

    def createGraph(self,rootNode):
        self.contactDistance = 0.1
        self.alarmDistance = 0.5
        self.exportVTK = 1    
        self.motionType = 1		# 0:  prescribed displacements of the center, 1: constant force

        self.createGenericComponents(rootNode)

        self.cells=[]   
        #self.createCellOEX319(rootNode)
        self.createCellOEX343_3(rootNode)
        

        # floor = rootNode.createChild('floor')
        # floor.createObject('RegularGrid', nx="2", ny="2", nz="2", xmin="-0.3", xmax="0.3",  ymin="-0.089", ymax="-0.091", zmin="-0.3", zmax="0.3")
        # floor.createObject('MechanicalObject', template="Vec3d")
        # floor.createObject('Triangle',simulated="false", bothSide="true", contactFriction="0.00", color="1 0 0 1")
        # floor.createObject('Line', simulated="false", bothSide="true", contactFriction="0.0")
        # floor.createObject('Point', simulated="false", bothSide="true", contactFriction="0.0")
    
        return 0;

    def createCellOEX343_3(self, rootNode):
        numCells=3
        trans3d = [[-64, -64, -64], [-48, -64, -64], [-64, -48, -64]]
        filoColors = [[0.1, 0.7, 0.3, 1], [0.3, 0.3, 1, 1], [0.1, 0.3, 0.3, 1]]
        bodyColors=[[0, 0.8, 0.2, 0.2], [0, 0, 1, 0.2], [0.3, 0.7, 0.2, 0.2] ]
        scale=1
        self.filopodiaConstFFScale = 0.000
        outPrefix='storeGeometry343_3cells/'
        fldSamplingDistance = 0.5
        cldMovingBoxes=[ [-1.5, -1.5, -1.5, 1.5, 1.5, 1.5], [-1.5+16, -1.5, -1.5, 1.5+16, 1.5, 1.5], [-1.5, -1.5+16, -1.5, 1.5, 1.5+16, 1.5]]
        cldMotionStiffnesses = [1e10, 1e10, 1e10]
        self.cldTraj = []
        self.cldTraj.append([ [0,  0.030, 0.000, 0.000,  0.0], [1.3, 0.000, 0.010, 0.000, 0.005], [3.0, -0.04, 0.040, 0.000, 0.0], [3.8, 0.04, 0.04, 0.0, 0.0], [5.0, -0.04, -0.0, 0.0, 0.0]])
        self.cldTraj.append([ [0, -0.030, 0.000, 0.000,  0.0], [1.3,-0.030,-0.020, 0.000, 0.006], [3.0, -0.06, 0.060, 0.000, 0.0], [3.8, 0.00, 0.01, 0.0, 0.0], [5.0, -0.04, -0.04, 0.0, 0.0]])
        self.cldTraj.append([ [0,  0.000,-0.040, 0.000,  0.01],[1.8,-0.020,-0.030, 0.000, 0.000] ])

        self.cldFiloForces = []
        self.cldFiloForces.append([[0, 200.0, 0.0, 0.0 ,0.0 ,0.0, 0.0], [1.5, 200.0, 100.0, 0.0, 0.0, 0.0, 0.0], [2.5, -200.0, 200.0, 0.0, 0.0, 0.0, 0.0], [4.2, 200.0, 200.0, 0.0, 0.0, 0.0, 0.0], [5.2, -200.0, 50.0, 0.0, 0.0, 0.0, 0.0] ])
        self.cldFiloForces.append([[0, -200.0, 0.0, 0.0 ,0.0 ,0.0, 0.0], [1.5, -200.0, -100.0, 0.0, 0.0, 0.0, 0.0], [2.5, -200.0, 200.0, 0.0, 0.0, 0.0, 0.0], [4.8, -200.0, -200.0, 0.0, 0.0, 0.0, 0.0]])        
        self.cldFiloForces.append([[0, -200.0, 0.0, 0.0 ,0.0 ,0.0, 0.0], [1.5, -200.0, -100.0, 0.0, 0.0, 0.0, 0.0], [2.5, -200.0, 200.0, 0.0, 0.0, 0.0, 0.0], [4.8, -200.0, -200.0, 0.0, 0.0, 0.0, 0.0]]) 
        
        
        for cellID in range(0,numCells):
            cellNode = rootNode.createChild('cell'+str(cellID))
            
            flds=[]            
            fldGen=FiloData()
            fldGen.femRadius=str(0.3*scale)
            fldGen.visRadius=str(0.2*scale)
            fldGen.youngModulus = 5e5
            fldGen.poissonRatio = 0.4
            fldGen.totalMass = 1e1
            fldGen.color=filoColors[cellID]
            fldGen.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            fldGen.trans3d=trans3d[cellID]
            #fldFileID=['01', '02', '03', '04', '05', '06', '07', '08', '09', '11', '12']
            fldFileID=['02', '08', '09', '13', '14', '15', '16', '17', '18', '19', '20', '21']
            #fldFileID=['01', '02', '03', '05', '06', '07', '08', '09', '11', '13', '14', '15', '17', '19', '20', '21', '22']

            for fli in xrange(0,len(fldFileID)):
                fld = copy.copy(fldGen)
                fld.id=fli
                fld.file='../data/overExp_ID343/ID343_t019_fTree'+str(fldFileID[fli])+'.vtk'
                fld.outFile = outPrefix + 'C'+str(cellID)+'_fTree' + str(fldFileID[fli]) + '.vtk'
                
                fld.forceIndices = []
                fld.forces=[]
                fld.cellID = cellID                

                fld.samplingDistance = fldSamplingDistance
                fld.perturbedForce = [0, 0, 0, 0, 0, 0]
                
                self.createFilop(cellNode,fld)
                flds.append(fld)
                fli = fli+1
            
            cld=CellData()
            cld.id=cellID
            cld.massDensity=20 # 0 #0
            cld.youngModulus = 1e4 # 1e6
            cld.poissonRatio = 0.45
            cld.rmass = 0.1
            cld.rstiffness = 0.1


            cld.bodyFile='../data/overExp_ID343/ID343_t019_nucleus.vtk'
            cld.surfaceFile='../data/overExp_ID343/ID343_t019_nucleusSurf.vtk'            
            cld.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            cld.trans3d=trans3d[cellID]            
            cld.color=bodyColors[cellID]
            cld.filopodia = flds
            cld.bodyOutFile = outPrefix + 'C'+str(cellID)+'_nucleus.vtk'
            cld.surfaceOutFile = outPrefix + 'C'+str(cellID)+'_nucleusSurf.vtk'
            cld.movingBox = cldMovingBoxes[cellID]
            cld.motionStiffness = cldMotionStiffnesses[cellID]

                    
            self.createCellBody(cellNode,cld)
            self.cells.append(cld)

        return 0


    def createGenericComponents(self, rootNode):
        rootNode.findData('dt').value="0.01"
        rootNode.findData('gravity').value="0 0 0"     
        rootNode.createObject('VisualStyle', displayFlags='showCollisionModels hideForceFields hideBehaviorModels hideWireframe')

        rootNode.createObject('FreeMotionAnimationLoop')
        rootNode.createObject('GenericConstraintSolver', maxIterations='1000', tolerance='1e-6', printLog='0', allVerified='0')

        rootNode.createObject('CollisionPipeline', depth="6", verbose="0", draw="0")
        rootNode.createObject('BruteForceDetection', name="N2")
        rootNode.createObject('LocalMinDistance', name="Proximity",  alarmDistance=self.alarmDistance, contactDistance=self.contactDistance,  angleCone='90.0', filterIntersection='0')
        rootNode.createObject('DefaultContactManager', name="Response", response="FrictionContact", responseParams='mu=0')

        self.timeStep = 0


    def createCellBody(self, cellNode, cld):
        #cellNode/phantomCell: to move the actual cell
        cellNode.createObject('MeshVTKLoader', name='vloader', filename=cld.bodyFile, translation=cld.trans3d, scale3d=cld.scale3d)
        cellNode.createObject('MeshVTKLoader', name='sloader', filename=cld.surfaceFile, flipNormals='0', translation=cld.trans3d, scale3d=cld.scale3d)

        if self.motionType == 1:
	        bodyAux = cellNode.createChild('bodyAux')
	        bodyAux.createObject('EulerImplicitSolver', firstOrder='0', rayleighStiffness='0.1', rayleighMass='0.1')
	        bodyAux.createObject('CGLinearSolver', tolerance='1e-10', threshold='1e-10')
	        bodyAux.createObject('GenerateRigidMass', name='massEngine', template="Rigid", position='@../sloader.position', triangles="@../sloader.triangles")

	        #self.phantBodyCenters.append(bodyAux.createObject('MechanicalObject', name='mstate', translation="@massEngine.massCenter", template="Rigid", showObject="0", showObjectScale="0.01", showIndices='1'))
	        cld.bodyCenterMO = bodyAux.createObject('MechanicalObject', name='mstate', translation="@massEngine.massCenter", template="Rigid", showObject='0', showObjectScale="0.5", showIndices='0')
	        # bodyAux.createObject('UniformMass', name='dummyMass', totalmass='1')

	        phantCellVol = bodyAux.createChild('volume')
	        phantCellVol.createObject('TetrahedronSetTopologyContainer', src='@../../vloader', name='Container', tags=' ')
	        phantCellVol.createObject('MechanicalObject', name='mstate', src='@../../vloader', showObject='0')
	        phantCellVol.createObject('RigidMapping', globalToLocalCoords='1')
	        # phantCellVol.createObject('Line')
        
        # cellNode/body
        body = cellNode.createChild('body')
        body.createObject('EulerImplicitSolver', rayleighStiffness=cld.rmass, rayleighMass=cld.rstiffness)
        # body.createObject('SparseLDLSolver')  
        body.createObject('SparsePARDISOSolver', name='lsolver', exportDataToFolder='', symmetric="1", pardisoSchurComplement='0')        
        
        body.createObject('TetrahedronSetTopologyContainer', src='@../vloader', name='Container', tags=' ')
        body.createObject('TetrahedronSetTopologyModifier', name='Modifier')
        body.createObject('TetrahedronSetTopologyAlgorithms', name='TopoAlgo', template='Vec3d')
        body.createObject('TetrahedronSetGeometryAlgorithms', name='GeomAlgo', template='Vec3d')
        body.createObject('MechanicalObject', name='mstate')
        # body.createObject('UniformMass', totalmass='1')
        body.createObject('MeshMatrixMass', name='mass', massDensity=cld.massDensity)
        body.createObject('TetrahedronFEMForceField', poissonRatio=cld.poissonRatio, youngModulus=cld.youngModulus, method='large', name='FEM', drawHeterogeneousTetra='1')
        body.createObject('BoxROI', box=cld.movingBox, name='fixedBox', drawBoxes='1')
        if self.motionType == 1:
        	body.createObject('RestShapeSpringsForceField', points='@fixedBox.indices', stiffness=cld.motionStiffness, external_rest_shape='@../bodyAux/volume/mstate', external_points='@fixedBox.indices')
        	# body.createObject('ExtendedRestShapeSpringForceField', points='@fixedBox.indices', stiffness=cld.motionStiffness, external_rest_shape='../bodyAux/volume/mstate', external_points='@fixedBox.indices')
        #elif self.motionType == 1:
        #	print 'NA'


        # body.createObject('ExtendedRestShapeSpringForceField', points='509 520 537', stiffness='1e2', external_rest_shape='../bodyAux/volume/mstate', external_points='509 520 537')
        body.createObject('PardisoConstraintCorrection', solverName='lsolver', schurSolverName='lsolver')        
        # body.createObject('LinearSolverConstraintCorrection')
        if self.exportVTK:
            body.createObject('VTKExporter', listening='1', filename=cld.bodyOutFile, XMLformat='0', edges='0', triangles='0', quads='0', tetras='1', exportEveryNumberOfSteps='1', exportAtBegin='1', exportAtEnd='0')


        # cellNode/body/surface
        surface = body.createChild('surface')        
        surface.createObject('TriangleSetTopologyContainer', position='@../../sloader.position', name='TriangleContainer', triangles='@../../sloader.triangles')
        surface.createObject('TriangleSetTopologyModifier', name='Modifier')
        surface.createObject('MechanicalObject', showIndices='false', name='mstate')
        surface.createObject('Triangle', color=cld.color, group=cld.id)
        surface.createObject('Line', color=cld.color, group=cld.id)
        surface.createObject('Point', color=cld.color, group=cld.id)
        surface.createObject('BarycentricMapping', name='bpmapping')
        if self.exportVTK:
            surface.createObject('VTKExporter', listening='1', filename=cld.surfaceOutFile, XMLformat='0', edges='0', triangles='1', quads='0', tetras='0', exportEveryNumberOfSteps='1', exportAtBegin='1', exportAtEnd='0')
        # surface.createObject('VectorDataIO', template='StdVectorTypes<Vec<3, double>, Vec<3, double>, double> >', name='ioPressCoords', inputFile=cld.pressureCoordFile,  readOnInit='1')
        # surface.createObject('VectorDataIO', template='StdVectorTypes<Vec<1, double>, Vec<1, double>, double> >', name='ioPressures', 
            # inputFileTimeMask=cld.pressureValueFile, readOnBeginStep=cld.pressureReadStep)
        # surface.createObject('NucleusPressureForceField', name='ppff', pressurePoints='@ioPressCoords.inputVector', pressureInBoundaryPoints='@ioPressures.inputVector', 
            # drawForceScale='1e-2', drawBoundaryPoints='0', pressureMultCoeff=cld.pressureMagnitude)
        #surface.createObject('VTKExporter', listening='1', filename=self.pDM.fNameNucleusSurfOut, XMLformat='0', edges='0', triangles='1', quads='0', tetras='1', exportEveryNumberOfSteps='0', exportAtEnd='1')

        # body.createObject('VTKExporter', exportEveryNumberOfSteps='0', exportAtBegin='1', exportAtEnd='0', edges='0', tetras='1', listening='1', XMLformat='0', filename='test.vtk')       

        #print '!!!!!!!!!!!!!!!!!!!!!!!!!!!: ',cld.filopodia

        # dummy = cellNode.createChild('dummy')
        # dummy.createObject('OglModel')

        for fli in xrange(0,len(cld.filopodia)):        
           filoRoot = body.createChild('filoRoot'+str(fli))
           filoRoot.createObject('MechanicalObject', position='@../../filo'+str(fli)+'/tree.treeRoot', name='MO', template='Rigid',showObjectScale="0.01", showIndices='0', showObject='0')
           filoRoot.createObject('BarycentricMapping')
           cellNode.createObject('BilateralInteractionConstraint', object1='@filo'+str(fli)+'/MO', object2='@body/filoRoot'+str(fli)+'/MO', second_point='0', first_point='0', template='Rigid')

        return 0

    def createFilop(self, cellNode, fld):
        fl = cellNode.createChild('filo'+str(fld.id))
        fl.activated = '1'
        fl.createObject('EulerImplicitSolver', firstOrder='0', rayleighStiffness='0.1', rayleighMass='0.1')
        fl.createObject('BTDLinearSolver')
        fl.createObject('MeshVTKLoader', filename=fld.file, name='loader', translation=fld.trans3d, scale3d=fld.scale3d)
        fl.createObject('BezierSegmentTree', name='tree', showCurveDiscretization='50', verbose='0', printLog='1', fittingError='-0.001', inputPositions='@loader.position', 
            maxSegments='1000', collisionOffset=2*self.contactDistance, exportBifurcationPairs='1', pointSamplingDistance=fld.samplingDistance, subdivideBranchIntoSegments='0', 
            inputPointDataSegmentIDs='@loader.pointDataSegmentID', showSampledPointsSize='0.0', showTangentSize='0.0',  showCurve='0', 
            showControlPointSize='0',  truncateOpenSegments='1.0',  showInputPointSize='0', cpointsIDFontSize='0.0')
        fl.createObject('Mesh', position='@tree.position', edges='@tree.edges', name='SkeletTopo')
        fl.createObject('MechanicalObject', position='@tree.position', showIndicesScale='5e-6', name='MO', template='Rigid', showIndices='0', showObject='0')
        fl.createObject('BeamFEMForceField', poissonRatio=fld.poissonRatio, radius=fld.femRadius, name='BeamFEM', useSymmetricAssembly='1', youngModulus=fld.youngModulus)

        if self.motionType == 1:        	
        	fl.createObject('BoxROI', box='-1000 -1000 -1000 1000 1000 1000', name='fixedBox', drawBoxes='1')
        	fld.constantForce = fl.createObject('ConstantForceField', indices='@fixedBox.indices', force='0 0 0 0 0 0', arrowSizeCoef=self.filopodiaConstFFScale)

        if fld.forceIndices:
            if fld.forceIndices=='file':
                fl.createObject('VectorDataToPointData', importedValues='@loader.forces', scaleFactor='0.001', name='forceConvert')
                fl.createObject('ConstantForceField', indices='@forceConvert.pointIndices', forces='@forceConvert.pointValues', arrowSizeCoef='1')
            else:                
                fl.createObject('ConstantForceField', indices=fld.forceIndices, force=fld.forces, arrowSizeCoef='0.001')

        fl.createObject('UniformMass', showAxisSizeFactor='0.001', totalmass=fld.totalMass)
        # fila1.createObject('DiagonalMass', template='Rigid', name='mass', massDensity='1000', mstate="@MO")
        fl.createObject('LinearSolverConstraintCorrection')
        # fila1.createObject('Line')

        #if self.motionType == 1:
       # 	ConstantForce
        
                

        flColl = fl.createChild('collision')
        flColl.createObject('MechanicalObject', name="MO", position="@../tree.collisionPosition")
        flColl.createObject('Mesh', position='@MO.position', edges='@../tree.collisionEdges')        
        flColl.createObject('Line', group=fld.cellID+1000, color='1 1 1 0')
        flColl.createObject('Point', group=fld.cellID+1000, color='1 1 1 0')
        flColl.createObject('BeamLinearMapping', localCoord='0')

                    
        #cellNode/fila1/VisuThread
        flVis = fl.createChild('visualization')
        flVis.createObject('OglModel', color=fld.color, name='Visual')
        flVis.createObject('QuadSetTopologyContainer', name='Container')
        flVis.createObject('QuadSetTopologyModifier', name='Modifier')
        flVis.createObject('Edge2QuadTopologicalMapping', nbPointsOnEachCircle='10', radius=fld.radius)
        flVis.createObject('TubularMapping', nbPointsOnEachCircle='10', radius=fld.visRadius)
        ## VisuThread.createObject('VTKExporter', exportEveryNumberOfSteps='0', quads='1', XMLformat='0', exportAtEnd='1', filename='storeGeometry/ID1_filoTree01quad.vtk', edges='0', listening='1', position='@Visual.position')
        
        flExp = fl.createChild('export')
        flExp.createObject('BezierSegmentTree', name='tree', showCurveDiscretization='50', verbose='0', printLog='1', fittingError='-0.001', inputPositions='@../loader.position', 
            maxSegments='1000', exportBifurcationPairs='0', pointSamplingDistance='0.0', subdivideBranchIntoSegments='0',  collisionOffset='-1',
            inputPointDataSegmentIDs='@loader.pointDataSegmentID', showSampledPointsSize='0.0', showTangentSize='0.0',  showCurve='0', 
            showControlPointSize='0',  truncateOpenSegments='1.0',  showInputPointSize='0', cpointsIDFontSize='0.0')
        flExp.createObject('Mesh', position='@tree.collisionPosition', edges='@tree.collisionEdges', name='SkeletTopo')
        flExp.createObject('MechanicalObject', position='@tree.collisionPosition', showIndicesScale='5e-6', name='MO', template='Vec3d', showIndices='0', showObject='1')

        if self.exportVTK:
            flExp.createObject('VTKExporter', exportEveryNumberOfSteps='1', exportAtBegin='1', exportAtEnd='0', listening='1', XMLformat='0', filename=fld.outFile,
                pointsDataFields='../loader.radiusByPoint ../loader.pointDataSegmentID')
        # flExp.createObject('VisualStyle', displayFlags='showCollisionModels hideForceFields showBehaviorModels hideWireframe')
        # flExp.createObject('ShowSpheres', position='@MO.position', radius='0.03', color="1 1 0 1")
        flExp.createObject('BeamLinearMapping', localCoord='0')



        return 0
    

    def onBeginAnimationStep(self, deltaTime):        
    	# move the cell 1
        #print self.phantBodyCenters
        actualTime = deltaTime * self.timeStep

        #if self.motionType == 0:
       	self.updateCellCenter(actualTime)   
       	#elif self.motionType == 1:     
       	self.updateFiloForces(actualTime)

        self.timeStep = self.timeStep+1

        return 0;

    def updateFiloForces(self, actualTime):
    	for cli in xrange(0,len(self.cells)):            
            filoForce = self.cldFiloForces[cli]            

            force = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

            randTime = self.timeStep % 20

            if (actualTime >= filoForce[-1][0]):
                force = [filoForce[-1][1], filoForce[-1][2], filoForce[-1][3], filoForce[-1][4], filoForce[-1][5], filoForce[-1][6]]
            else:
                for tp in xrange(1,len(filoForce)):
                    prevTime = filoForce[tp-1][0]
                    nextTime = filoForce[tp][0]
                    #print 'Time Segment:',prevTime, '->', nextTime
                    if (actualTime >= prevTime and actualTime < nextTime):
                    	force = [filoForce[tp-1][1], filoForce[tp-1][2], filoForce[tp-1][3], filoForce[tp-1][4], filoForce[tp-1][5], filoForce[tp-1][6]]                        
                        break
            print 'Time step: ',actualTime,' force: ', force



            for fli in xrange(0, len(self.cells[cli].filopodia)):
            	if randTime == 0:
            		f3 = force[0:3]
            		f3o1 = []

            		if f3[0] != f3[1]:
            			f3o1 = [-f3[1], f3[0], 0]
            		elif f3[0] != f3[2]:
            			f3o1 = [-f3[2], 0, -f3[0]]
            		elif f3[1] != f3[2]:
            			f3o1 = [0.0, -f3[0], f3[1]]

            		f3n = f3 / np.linalg.norm(f3)
            		f3o1n = f3o1 / np.linalg.norm(f3o1)
            		f3o2 = np.cross(f3n, f3o1n)
            		f3o2n = f3o2 / np.linalg.norm(f3o2)

            		r1 = 140*random.random()-70
            		r2 = 140*random.random()-70

            		pert = np.add(np.multiply(r1, f3o1n), np.multiply(r2, f3o2n))
            		pForce = np.add(f3, pert) + [0,0,0]            		
            		self.cells[cli].filopodia[fli].perturbedForce = [pForce[0], pForce[1], pForce[2], 0,0,0]
            		

            	print 'Force in filo ', fli,' ', self.cells[cli].filopodia[fli].perturbedForce
            	self.cells[cli].filopodia[fli].constantForce.findData("force").value = self.cells[cli].filopodia[fli].perturbedForce
            	#print 'Setting cell', cli, 'filop ', fli



    def updateCellCenter(self, actualTime):
    	for cli in xrange(0,len(self.cells)):            
            trajectory = self.cldTraj[cli]            

            dx = [0.0, 0.0, 0.0]
            da = 0.0

            if (actualTime >= trajectory[-1][0]):
                dx = [trajectory[-1][1], trajectory[-1][2], trajectory[-1][3]]
                da = trajectory[-1][4]
            else:
                for tp in xrange(1,len(trajectory)):
                    prevTime = trajectory[tp-1][0]
                    nextTime = trajectory[tp][0]
                    #print 'Time Segment:',prevTime, '->', nextTime
                    if (actualTime >= prevTime and actualTime < nextTime):
                        dx = [trajectory[tp-1][1], trajectory[tp-1][2], trajectory[tp-1][3]]
                        da = trajectory[tp-1][4]
                        break


            
            print 'Time step:',actualTime,' using dx: ', dx

            position = self.cells[cli].bodyCenterMO.findData("position").value
            # add 1 mm in x direction
            for d in xrange(0,3):
                position[0][d] = position[0][d] + dx[d]
                #position[0][d] = position[0][d] + self.cells[cli].centerTranslation[d]

            # rotate: 
            # convert actual quaternion to axis-angle
            qx = position[0][3]
            qy = position[0][4]
            qz = position[0][5]
            qw = position[0][6]

            angle = 2 * math.acos(qw)

            den = math.sqrt(1-qw*qw)

            #print 'den: ', den

            if den < 1e-8:
                ax = 1
                ay = 0
                az = 0
            else:
                ax = qx / den
                ay = qy / den
                az = qz / den
            #print "Actual angle: %f axis: %f %f %f" % (angle, ax, ay, az)

            #add something small to the angle, it's in radians!
            #angle = angle + self.cells[cli].centerRotationAngle
            angle = angle + da
            if angle > 2*np.pi:
                angle = 0

            #convert back to quaternion
            qx = ax * math.sin(angle/2)
            qy = ay * math.sin(angle/2)
            qz = az * math.sin(angle/2)
            qw = math.cos(angle/2)
            #print "New angle: %f New quaternion: %f %f %f %f\n" % (angle, qx, qy, qz, qw)
            position[0][3] = qx
            position[0][4] = qy
            position[0][5] = qz
            position[0][6] = qw
            
            # apply the new position&orientation to the rigid point
            self.cells[cli].bodyCenterMO.findData("position").value = position


    def onKeyPressed(self, c):
        return 0;

    def onKeyReleased(self, c):
        return 0;

    def onLoaded(self, node):
        return 0;

    def onMouseButtonLeft(self, mouseX,mouseY,isPressed):
        return 0;

    def onMouseButtonRight(self, mouseX,mouseY,isPressed):
        return 0;

    def onMouseButtonMiddle(self, mouseX,mouseY,isPressed):
        return 0;

    def onMouseWheel(self, mouseX,mouseY,wheelDelta):
        return 0;

    def onGUIEvent(self, strControlID,valueName,strValue):
        return 0;

    

    def onEndAnimationStep(self, deltaTime):        
        return 0;

    def onScriptEvent(self, senderNode, eventName,data):
        return 0;

    def initGraph(self, node):
        return 0;

    def bwdInitGraph(self, node):
        return 0;

    def storeResetState(self):
        return 0;

    def reset(self):
        return 0;

    def cleanup(self):
        return 0;


    #configuration of interaction of two 343 cells with Python trajectories
    def createCellOEX343_2(self, rootNode):           
        trans3d = [[-64, -64, -64], [-48, -64, -64]]        
        filoColors = [[1, 0.3, 1, 1], [0.3, 0.3, 1, 1]]
        bodyColors=[[1, 0, 1, 0.2],[0, 0, 1, 0.2]]            
        centerTransl = [ [0.03, 0, 0], [-0.03, 0, 0] ]
        rotAngles = [0.000, 0.006]        
        scale=1
        outPrefix='storeGeometry343A/'
        fldSamplingDistance = 0.5
        cldMovingBoxes=[ [-1.5, -1.5, -1.5, 1.5, 1.5, 1.5], [-1.5+16, -1.5, -1.5, 1.5+16, 1.5, 1.5]]
        cldMotionStiffnesses = [1e10, 1e10]
        self.cldTraj = []
        self.cldTraj.append([ [0,  0.030, 0.000, 0.000,  0.0], [1.3, 0.000, 0.010, 0.000, 0.005], [3.0, -0.04, 0.040, 0.000, 0.0], [3.8, 0.04, 0.04, 0.0, 0.0], [5.0, -0.04, -0.04, 0.0, 0.0]])
        self.cldTraj.append([ [0, -0.030, 0.000, 0.000,  0.0], [1.3,-0.030,-0.020, 0.000, 0.006], [3.0, -0.06, 0.060, 0.000, 0.0], [3.8, 0.00, 0.01, 0.0, 0.0], [5.0, -0.04, -0.04, 0.0, 0.0]])
        
        for cellID in range(0,2):
            cellNode = rootNode.createChild('cell'+str(cellID))
            
            flds=[]            
            fldGen=FiloData()
            fldGen.femRadius=str(0.3*scale)
            fldGen.visRadius=str(0.2*scale)
            fldGen.youngModulus = 5e5
            fldGen.poissonRatio = 0.4
            fldGen.totalMass = 1e2
            fldGen.color=filoColors[cellID]
            fldGen.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            fldGen.trans3d=trans3d[cellID]
            #fldFileID=['01', '02', '03', '04', '05', '06', '07', '08', '09', '11', '12']
            fldFileID=['02', '08', '09', '13', '14', '15', '16', '17', '18', '19', '20', '21']
            #fldFileID=['01', '02', '03', '05', '06', '07', '08', '09', '11', '13', '14', '15', '17', '19', '20', '21', '22']

            for fli in xrange(0,len(fldFileID)):
                fld = copy.copy(fldGen)
                fld.id=fli
                fld.file='../data/overExp_ID343/ID343_t019_fTree'+str(fldFileID[fli])+'.vtk'
                fld.outFile = outPrefix + 'C'+str(cellID)+'_fTree' + str(fldFileID[fli]) + '.vtk'
                
                fld.forceIndices = [0, 1, 2]
                fld.forces='[0 0 0 0 0 0]'
                fld.cellID = cellID

                fld.samplingDistance = fldSamplingDistance
                
                self.createFilop(cellNode,fld)
                flds.append(fld)
                fli = fli+1
            
            cld=CellData()
            cld.id=cellID
            cld.massDensity=1000
            cld.youngModulus = 1e6
            cld.poissonRatio = 0.45
            cld.rmass = 0.1
            cld.rstiffness = 0.


            cld.bodyFile='../data/overExp_ID343/ID343_t019_nucleus.vtk'
            cld.surfaceFile='../data/overExp_ID343/ID343_t019_nucleusSurf.vtk'            
            cld.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            cld.trans3d=trans3d[cellID]            
            cld.color=bodyColors[cellID]
            cld.centerTranslation=centerTransl[cellID]
            cld.centerRotationAngle=rotAngles[cellID]            
            cld.filopodia = flds
            cld.bodyOutFile = outPrefix + 'C'+str(cellID)+'_nucleus.vtk'
            cld.surfaceOutFile = outPrefix + 'C'+str(cellID)+'_nucleusSurf.vtk'
            cld.movingBox = cldMovingBoxes[cellID]
            cld.motionStiffness = cldMotionStiffnesses[cellID]

                    
            self.createCellBody(cellNode,cld)
            self.cells.append(cld)

        return 0

    #configuration of 2 cells (only a simple interaction, probably not compatible anymore with actual trajectories)
    def createCellOEX319(self, rootNode):
        self.contactDistance = 0.1
        self.alarmDistance = 0.5
        trans3d = ['-64 -64 -64', '-45 -64 -64']        
        filoColors = ['1 1 0 1', '0 1 0 1']
        bodyColors=['1 0 1 0.2','1 1 0 0.2']            
        centerTransl = [ [0.005, 0, 0], [-0.03, 0, 0] ]
        rotAngles = [0.000, 0.006]        
        scale=1
        outPrefix='storeGeometry4/'
        fldSamplingDistance = 0.2

        for cellID in range(0,2):            
            cellNode = rootNode.createChild('cell'+str(cellID))
            
            flds=[]            
            fldGen=FiloData()
            fldGen.radius=str(0.25*scale)
            fldGen.youngModulus = 5e5
            fldGen.poissonRatio = 0.4
            fldGen.totalMass = 5
            fldGen.color=filoColors[cellID]
            fldGen.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            fldGen.trans3d=trans3d[cellID]
            fldFileID=['01', '02', '03', '04', '05', '06', '07', '08', '09', '11', '12']

            for fli in xrange(0,len(fldFileID)):
                fld = copy.copy(fldGen)
                fld.id=fli
                fld.file='../data/overExp_ID319/ID319_t000_fTree'+str(fldFileID[fli])+'ok.vtk'
                fld.outFile = outPrefix + 'C'+str(cellID)+'_fTree' + str(fldFileID[fli]) + '.vtk'
                
                fld.forceIndices = 'file'
                fld.cellID = cellID

                fld.samplingDistance = fldSamplingDistance
                
                self.createFilop(cellNode,fld)
                flds.append(fld)
                fli = fli+1
            
            cld=CellData()
            cld.id=cellID
            cld.bodyFile='../data/overExp_ID319/ID319_t000_nucleus.vtk'
            cld.surfaceFile='../data/overExp_ID319/ID319_t000_nucleusSurf.vtk'
            cld.pressureCoordFile = '../data/overExp_ID319/pressures/coord.txt'
            cld.pressureValueFile = '../data/overExp_ID319/pressures/pressures_T%03d.txt'
            cld.pressureMagnitude=0
            cld.pressureReadStep=100
            cld.scale3d=str(scale)+' '+str(scale)+' '+str(scale)
            cld.trans3d=trans3d[cellID]
            cld.youngModulus = 1e5
            cld.poissonRatio = 0.45
            cld.color=bodyColors[cellID]
            cld.centerTranslation=centerTransl[cellID]
            cld.centerRotationAngle=rotAngles[cellID]            
            cld.filopodia = flds
            cld.bodyOutFile = outPrefix + 'C'+str(cellID)+'_nucleus.vtk'
            cld.surfaceOutFile = outPrefix + 'C'+str(cellID)+'_nucleusSurf.vtk'
                    
            self.createCellBody(cellNode,cld)
            self.cells.append(cld)

        return 0


    # def createGraph(self,rootNode):
    #     self.createGenericComponents(rootNode)

    #     cells=['cell1','cell2']
    #     self.phantBodyCenters = []
    #     cellID=0
    #     filoIDs=[]

                
    #     cellID=cellID+1
    #     filoID=0        
    #     cellNode1 = rootNode.createChild('cell1')
    #     forceIndices='5 10'
    #     forces='0 0 0.002 0 0 0   0 0.003 0.005 0 0 0'
    #     filoID=filoID+1
    #     self.createFilop(cellNode1, cellID, filoID, '../data/ID1_FiloTrees/ID1_t001_fTree01.vtk', '0 0 0', forceIndices, forces, '1 1 0 1')        
    #     forceIndices = ''
    #     filoID=filoID+1
    #     self.createFilop(cellNode1, cellID, filoID, '../data/ID1_FiloTrees/ID1_t001_fTree02.vtk', '0 0 0', forceIndices, forces, '1 1 0 1')
    #     self.createCellBody(cellNode1, cellID, filoID, '../data/ID1_Nucleus/ID1_t001_nucleus2038.vtk', '../data/ID1_Nucleus/ID1_t001_nucleusSurf500.vtk', '0 0 0', '1 0 0 1')
    #     filoIDs.append(filoID)


    #     cellID=cellID+1
    #     filoID=0
    #     cellNode2 = rootNode.createChild('cell2')
    #     transl='-0.03 0.005 0.09'
    #     # forceIndices=''
    #     # forces=''
    #     # filoID=filoID+1
    #     # self.createFilop(cellNode2, cellID, filoID, '../data/ID1_FiloTrees/ID1_t001_fTree01.vtk', transl, forceIndices, forces, '0 1 1 1')
    #     forceIndices = 'file'
    #     filoID=filoID+1
    #     self.createFilop(cellNode2, cellID, filoID, '../data/ID1_FiloTrees/ID1_t001_fTree02.vtk', transl, forceIndices, forces, '0 1 1 1')
    #     self.createCellBody(cellNode2, cellID,  filoID, '../data/ID1_Nucleus/ID1_t001_nucleus2038.vtk', '../data/ID1_Nucleus/ID1_t001_nucleusSurf500.vtk', transl, '0.5 0.2 0.5 1.0')
    #     filoIDs.append(filoID)


    #     # floor = rootNode.createChild('floor')
    #     # floor.createObject('RegularGrid', nx="2", ny="2", nz="2", xmin="-0.3", xmax="0.3",  ymin="-0.089", ymax="-0.091", zmin="-0.3", zmax="0.3")
    #     # floor.createObject('MechanicalObject', template="Vec3d")
    #     # floor.createObject('Triangle',simulated="false", bothSide="true", contactFriction="0.00", color="1 0 0 1")
    #     # floor.createObject('Line', simulated="false", bothSide="true", contactFriction="0.0")
    #     # floor.createObject('Point', simulated="false", bothSide="true", contactFriction="0.0")
    
    #     return 0;        

